Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq fintype tuple finfun. 
Require Import finset rels heaps stmodel stsep hprop vcgen array fmap ordtype kvmaps.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Module HashTab. Section HashTab.
Variables (K : ordType) (V : Type) (buckets : KVmap.Sig K V) 
          (n : nat) (hash : K -> 'I_n).
Definition hashtab := {array 'I_n -> KVmap.tp buckets}.
Notation KVshape := (@KVmap.shape _ _ buckets).
Notation table := (table KVshape).

Definition shape x s := 
  [Rel h | exists tab, exists bucket, 
           (forall k, fnd k s = fnd k (bucket (hash k))) /\
           (forall i k, k \in supp (bucket i) -> hash k = i) /\
           h \In Array.shape x tab # sepit setT (table tab bucket)].

Lemma shapeD x s : Def (shape x s).
Proof. move=>xs s h [t][b][_][_][h1][h2]; tauto. Qed.
  
Lemma shape_invert : forall (x : hashtab), invertible (shape x).
Proof.
move=>x s1 s2 h1 h2 [t1][b1][SH1][S1] H1 [t2][b2][SH2][S2] H2 R.
have H: forall t, invertible (KVshape t) by apply: KVmap.shape_invert.
suff [L1 L2 L3]: [/\ t1 = t2, forall i, b1 i = b2 i & h1 = h2].
- by split=>//; apply: fmapP=>k; rewrite SH1 SH2 L2.
by apply: (table_invert (x:=x) H) R.
Qed.

Definition new_loopinv x := forall k, 
  STsep hashtab (fun h => k <= n /\ exists tab, h \In Array.shape x tab # 
                   sepit [set x:'I_n | x < k] (table tab (fun x:'I_n => nil K V)), 
                 fun y i m => exists x, y = Val x /\ shape x (nil K V) m). 

Program 
Definition new : STsep hashtab (emp, fun y i m => exists x, y = Val x /\ 
                                                  m \In shape x (nil K V)) := 
  Do (t <-- Array.new _ (KVmap.default buckets); 
      Fix (fun (loop : new_loopinv t) k => 
             Do (Match_dec (dec (k < n))
                  (fun pf => b <-- KVmap.new buckets; 
                             Array.write t (Ordinal pf) b;; 
                             loop k.+1)
                  (fun _ => ret t))) 0).
Next Obligation.
move: i H0 H1=>_ tab [h1][h2][->][H1]; case: dec=>[{H} pf [H2]|] /=; last first.
- case: ltnP H (eqn_leq k n)=>// _ -> /=; move/eqP=>-> _ [H2]; heval=>D.
  exists t; split=>//; exists tab; exists (fun _:'I_n => nil K V); hauto D.
  by apply: tableP2 H2=>// x; rewrite !in_set ltn_ord.
rewrite -(un0h (_ :+ _))=>D.
apply: bnd_do=>[|b i1' [_] [[<-]] H3|??[?][]|] {D} //.
rewrite unCA; apply: bnd_gh H1=>[[] h1' [H1'] _ D|??[]] //.
apply: cons0=>//; split=>//; exists (ffun_of [eta tab with Ordinal pf |-> b]).
hauto D; rewrite (sepitS (Ordinal pf)) in_set leqnn {1}/table ffunE /= eq_refl.
hauto D; apply: tableP2 H2=>{H1'} //.
- by case=>x ?; rewrite !in_set -val_eqE /= ltnS (leq_eqVlt x); case: ltngtP.
by move=>x _; rewrite in_set ffunE /=; case: eqP=>// ->; rewrite ltnn.
Qed.
Next Obligation.
rewrite -{1}(unh0 empty); apply: bnd_do=>[|a i1' [_][[<-]] H1 D|??[?][]|] //.
apply: cons0=>//; split=>//; exists [ffun : 'I_n => KVmap.default buckets].
by hauto D; rewrite (eq_sepit set0) ?sepit0.
Qed.

Definition free_loopinv x := forall k, 
  STsep unit (fun h => k <= n /\ exists t, exists b, h \In Array.shape x t # 
                       sepit [set x:'I_n | x >= k] (table t b), 
              fun y i m => y = Val tt /\ m = empty).

Program Definition free x : STsep unit (fun i => exists s, i \In shape x s, 
                                        fun y i m => y = Val tt /\ m = empty) :=
  Do (Fix (fun (loop : free_loopinv x) k => 
             Do (Match_dec (dec (k < n))
                  (fun pf => b <-- Array.read x (Ordinal pf);
                             KVmap.free b;;
                             loop k.+1)
                  (fun _ => Array.free x))) 0).
Next Obligation.
move: i H0 H1 H2=>h t b [h1][h2][->][H1][]; case: dec=>[{H} pf H2|]; last first.
- case: ltnP H (eqn_leq k n)=>// _ -> /=; move/eqP=>-> _.
  rewrite (eq_sepit set0) ?sepit0; first by case=>y p; rewrite in_set inE leqNgt p.
  move=>->; rewrite unh0=>D; apply: cons0=>/=; eauto.  
apply: bnd_gh (H1)=>[b' _ [[->]] <-|??[]] //; rewrite pf.
rewrite (sepitS (Ordinal pf)) in_set leqnn /table in H2.
move: H2=>[h3][h4][->][H2][H3] _; rewrite unCA.
apply: bnd_do=>[|[]_[_]->|??[]] //=; first by eauto.
rewrite un0h=>D; apply: cons0=>//; split=>//; exists t; exists b; hauto D.
apply: tableP2 H3=>//; move=>[y qf]; rewrite !in_set -val_eqE /=.
by case: eqP=>[->|]; rewrite ?ltnn // leq_eqVlt; case: ltngtP=>// ->.
Qed. 
Next Obligation. 
by apply: cons0=>//=; [case: H0=>t [b][_][_]; eauto| apply: shapeD H0].
Qed.

Program 
Definition insert x k v : STsep unit (fun i => exists s, i \In shape x s, 
                                      fun y i m => forall s, i \In shape x s -> 
                                        m \In shape x (ins k v s) /\ y = Val tt) :=
  Do (b <-- Array.read x (hash k);
      KVmap.insert b k v).
Next Obligation.
apply: (ghost H) H0=>s [t][b][H1][H2][h1][h2][->][H3][H4]. 
apply: bnd_gh (H3)=>[b' _ [[->]] <-|??[]] //; rewrite (sepitT1 (hash k)) in H4.
move: H4; rewrite /table=>[[h2']][h3][->][H4][H5] _; rewrite unCA. 
apply: val_gh H4=>[[] h2'' [H4'] _ D|??[]] //; split=>//.
set b'' := [eta b with (hash k) |-> ins k v (b (hash k))].
exists t; exists (b'' : 'I_n -> {fmap K -> V}); do !split. 
- move=>k'; rewrite /b'' /= fnd_ins.
  case: ifP=>T1; first by rewrite (eqP T1) eq_refl fnd_ins eq_refl.
  case: ifP=>T2 //; first by rewrite fnd_ins T1 -(eqP T2).
- move=>i' k'; rewrite /b'' /=; case: eqP=>[->|_]; last by apply: H2.
  by rewrite supp_ins inE /=; case: eqP=>[->|_] //=; last by apply: H2.
rewrite (sepitT1 (hash k)) unCA {1}/table /= eq_refl in D *; hauto D.
by apply: tableP H5=>// y; rewrite !in_set andbT /b'' /=; move/negbET=>->.
Qed. 

Program 
Definition remove x k : STsep unit (fun i => exists s, i \In shape x s, 
                                    fun y i m => forall s, i \In shape x s -> 
                                      m \In shape x (rem k s) /\ y = Val tt) :=
  Do (b <-- Array.read x (hash k);
      KVmap.remove b k). 
Next Obligation.
apply: (ghost H) H0=>s [t][b][H1][H2][h1][h2][->][H3][H4].
apply: bnd_gh (H3)=>[b' _ [[->]] <-|??[]] //; rewrite (sepitT1 (hash k)) in H4.
move: H4; rewrite /table=>[[h2']][h3][->][H4][H5] _; rewrite unCA.
apply: val_gh H4=>[[] h2'' [H4'] _ D|??[]] //; split=>//.
set b'' := [eta b with (hash k) |-> rem k (b (hash k))].
exists t; exists (b'' : 'I_n -> fmap K V); do !split.
- move=>k'; rewrite /b'' /= fnd_rem.
  case: ifP=>T1; first by rewrite (eqP T1) eq_refl fnd_rem eq_refl.
  case: ifP=>T2 //; first by rewrite fnd_rem T1 -(eqP T2).
- move=>i' k'; rewrite /b'' /=; case: eqP=>[->|_]; last by apply: H2.
  by rewrite supp_rem inE /=; case: eqP=>[->|_] //=; last by apply: H2.
rewrite (sepitT1 (hash k)) unCA {1}/table /= eq_refl in D *; hauto D. 
by apply: tableP H5=>// y; rewrite !in_set andbT /b'' /=; move/negbET=>->.
Qed. 

Program Definition lookup x k : 
  STsep (option V) (fun i => exists s, i \In shape x s, 
                    fun y i m => forall s, i \In shape x s -> 
                                   m \In shape x s /\ y = Val (fnd k s)) :=
  Do (b <-- Array.read x (hash k); 
      KVmap.lookup b k).
Next Obligation.
apply: (ghost H) H0=>s [t][b][H1][H2][h1][h2][->][H3][H4].
apply: bnd_gh (H3)=>[b' _ [[->]] <-|??[]] //; rewrite (sepitT1 (hash k)) in H4.
move: H4; rewrite /table=>[[h2']][h3][->][H4][H5] _; rewrite unCA.
apply: val_gh H4=>[a h2'' [H4'] [->] {D} D|??[]] //; eauto; rewrite H1; split=>//.
exists t; exists b; do !split=>//.
by rewrite (sepitT1 (hash k)) unCA in D *; hauto D.
Qed. 

Definition HashTab := KVmap.make (Array null) shapeD shape_invert new free insert remove lookup.

End HashTab. End HashTab.

Definition HT K V := HashTab.HashTab (AssocList.AssocList K V).