Require Import ssreflect ssrbool ssrnat eqtype.
Require Import rels hprop heaps stmodel stsep vcgen seq llist.
Set Implicit Arguments. 
Unset Strict Implicit. 
Import Prenex Implicits.

Definition stack (T : Type) := loc.
Definition EmptyStack := exn_from_nat 25.

Module Stack. Section Stack.
Variable T : Type. 
Notation stack := (stack T).

Definition shape s xs := [Rel h | exists p, exists h', def (s :-> p :+ h') /\ 
                                    s :-> p :+ h' = h /\ h' \In @lseq T p xs]. 

Lemma shape_inv : forall s, invert (shape s). 
Proof.
move=>s xs1 xs2 _ [p][h][D][<-] S [p2][h2][D2][].
by case/(pts_injSl D2)=>->->; apply: lseq_inv.
Qed.
  
Program Definition new : STsep stack (emp, fun y i h => exists v, y = Val v /\ 
                                                          shape v [::] h) :=
  Do (alloc null).
Next Obligation. by case: H=><-; apply: val_alloc=>// D; hauto D. Qed.

Program Definition free s : STsep unit (shape s [::], 
                                        fun y i h => y = Val tt /\ h = empty) :=
  Do (dealloc s).
Next Obligation. by move: H3 H1=>[->]->; heval. Qed.

Program 
Definition push s x : STsep unit (fun i => exists xs, shape s xs i, 
                                  fun y i h => forall xs, shape s xs i -> 
                                    y = Val tt /\ shape s (x :: xs) h) :=
  Do (hd <-- !s; 
      nd <-- Allocb x 2;
      nd .+ 1 ::= (hd : loc);;
      s ::= nd).
Next Obligation.
apply: (ghost1 H)=>[?|]; first by apply: shape_inv; hhauto.
by move: H2; heval=>nd; heval; rewrite un0h -(push nd)=>D; hauto D.
Qed.  

Program Definition pop s : 
  STsep T (fun i => exists xs, shape s xs i, 
           fun y i h => forall xs, shape s xs i -> shape s (behead xs) h /\
             match y with Val v => xs = v :: behead xs
                        | Exn e => e = EmptyStack /\ xs = [::] end) :=
  Do (hd <-- !s;
      If hd == null then Throw T EmptyStack
      else
        x <-- !hd;
        next <-- !(hd .+ 1);
        dealloc hd;; 
        dealloc hd .+ 1;;
        s ::= (next : loc);;
        ret x).
Next Obligation.
apply: (ghost1 H)=>[?|]; first by apply: shape_inv; hhauto.
move: H2; heval; case: eqP H4=>[->|N]; first by case/lseq_null=>->->; heval.
by case/(lseq_pos (introN (eqlocP _ _) N))=>?[?][?][->] _ <- ?; heval.
Qed.

End Stack. End Stack.