Require Import ssreflect ssrfun ssrbool ssrnat eqtype seq fintype finfun finset.
Require Import rels heaps paths.
Set Implicit Arguments.
Unset Strict Implicit. 
Import Prenex Implicits.

Ltac add_morphism_tactic := SetoidTactics.add_morphism_tactic.
Notation " R ===> R' " := (@Morphisms.respectful _ _ R R')
  (right associativity, at level 55) : signature_scope.

Definition star (p1 p2 : Rel heap) :=
  [Rel h | exists h1, exists h2, h = h1 :+ h2 /\ h1 \In p1 /\ h2 \In p2 /\ def h].

Definition emp := [Rel i | i = empty].
Definition this i := [Rel h : heap | i = h].
Definition pts A x (v : A) := [Rel h | locked (def (x :-> v) /\ x :-> v = h)].

Definition Def (p : Rel heap) := forall h, h \In p -> def h. 

Notation "p1 '#' p2"   := (star p1 p2) (at level 57, right associativity) : rel_scope.
Notation "x '-->' v"   := (pts x v) (at level 50) : rel_scope.

(* we will often need to swap the conjuncts in pts, when destructing *)
(* for that, we introduce a view, with a really short name *)

Lemma swp : forall A (v : A) x h, h \In x --> v <-> h = x :-> v /\ def h.
Proof. by move=>A v x h; split; rewrite InE /=; unlock; [case=>D <- | case=>-> D]. Qed. 

Lemma opn : forall A (v : A) x h, h \In x --> v <-> def (x :-> v) /\ x :-> v = h.
Proof. by move=>A v x h; split=>[|H]; rewrite InE /=; unlock. Qed.

Prenex Implicits swp opn.

Inductive defeq (p1 p2 : Rel heap) : Prop := 
  defeqI of (forall h, def h -> h \In p1 -> h \In p2) & 
            (forall h, def h -> h \In p2 -> h \In p1).

Notation "p1 <=> p2" := (defeq p1 p2) (at level 80, no associativity) : rel_scope.

Section DefeqEquivalence.

Lemma defeq_refl p : p <=> p. Proof. by move=>p; split. Qed.

Lemma defeq_sym p1 p2 : p1 <=> p2 -> p2 <=> p1.
Proof. by move=>p1 p2 [H1 H2]; split=>h D; [move/(H2 h D) | move/(H1 h D)]. Qed.

Lemma defeq_trans' p1 p2 p3 : p1 <=> p2 -> p2 <=> p3 -> p1 <=> p3.
Proof.
move=>p1 p2 p3 [H1 H2] [H3 H4]; split=>h D;
by [move/(H1 h D); move/(H3 h D) | move/(H4 h D); move/(H2 h D)].
Qed.

Definition defeq_trans p2 p1 p3 := @defeq_trans' p1 p2 p3.

Lemma reldef p1 p2 : p1 =r p2 -> p1 <=> p2.
Proof. by move=>p1 p2 [H1 H2]; split=>h _; [apply: H1 | apply: H2]. Qed.

Lemma defrel p1 p2 : p1 <=> p2 -> Def p1 -> Def p2 -> p1 =r p2.
Proof.
move=>p1 p2 [H1 H2] H3 H4; split=>h T.
- by apply: H1 (T); apply: H3 T.
by apply: H2 (T); apply: H4 T.
Qed.

End DefeqEquivalence.

Add Relation (Rel heap) defeq 
  reflexivity proved by defeq_refl
  symmetry proved by defeq_sym
  transitivity proved by defeq_trans' as defeq_rel. 

Add Morphism star with signature defeq ===> defeq ===> defeq as star_morph.
Proof.
move=>p1 q1 [H1 H2] p2 q2 [H3 H4]; split=>h _; move=>[h1][h2][->][T1][T2] D;
exists h1; exists h2; do !split=>//.
- by apply: H1 T1; apply: unDl D.
- by apply: H3 T2; apply: unDr D.
- by apply: H2 T1; apply: unDl D.
by apply: H4 T2; apply: unDr D.
Qed.

Add Morphism star with 
  signature (@releq heap) ===> (@releq heap) ===> (@releq heap) as star_releq_morph.
Proof.
move=>p1 q1 [H1 H2] p2 q2 [H3 H4]; split=>h; move=>[h1][h2][->][T1][T2] D;
exists h1; exists h2; do !split=>//.
- by apply: H1 T1.
- by apply: H3 T2. 
- by apply: H2 T1. 
by apply: H4 T2.
Qed.  

Add Morphism defeq with 
  signature (@releq heap) ===> (@releq heap) ===> iff as defeq_morph.
Proof.
move=>p1 q1 [H1 H2] p2 q2 [H3 H4]; split; case=>T1 T2; split=>h D.
- by move/H2; move/(T1 _ D); move/H3.
- by move/H4; move/(T2 _ D); move/H1.
- by move/H1; move/(T1 _ D); move/H4.
by move/H3; move/(T2 _ D); move/H2.
Qed.
  
Section BasicProperties.

Lemma starC p1 p2 : p1 # p2 =r p2 # p1.
Proof. 
move=>p1 p2; split=>h;
by move=>[h1][h2][->][H1][H2] D; exists h2; exists h1; rewrite {1}unC.
Qed.

Lemma starp0 p : p # emp <=> p.
Proof.
move=>p; split=>[h _ [h1][h2][->][H1][->]| h D H]; first by rewrite unh0.
by exists h; exists empty; rewrite unh0.
Qed.

Lemma star0p p : emp # p <=> p.
Proof. by move=>p; rewrite starC starp0. Qed.

Lemma starCA p1 p2 p3 : p1 # p2 # p3 =r p2 # p1 # p3.
Proof.
move=>p1 p2 p3; split=>h; move=>[h1][_][->][H1][[h2]][h3][->][H2][H3] _; rewrite unCA=>D;
by do ![econstructor]=>//; apply: unDr D.
Qed.

Lemma starA p1 p2 p3 : (p1 # p2) # p3 =r p1 # p2 # p3.
Proof. by move=>p1 p2 p3; rewrite (starC p2) starCA starC. Qed.

Lemma starAC p1 p2 p3 : (p1 # p2) # p3 =r (p1 # p3) # p2.
Proof. by move=>p1 p2 p3; rewrite -2!(starC p3) starA. Qed.

Lemma starM : Monotone star.
Proof. 
move=>s1 r1 s2 r2 H1 H2 h; move=>[h1][h2][->][T1][T2] D.
by do ![econstructor]=>//; [apply: H1 | apply: H2].
Qed.

End BasicProperties.  

Section DefProperties.

Lemma def_empty : Def emp. Proof. by move=>h ->. Qed.

Lemma def_star p1 p2 : Def (p1 # p2).
Proof. by move=>p1 p2 h [h1][h2][->][_][_]. Qed.

End DefProperties.  

Hint Resolve def_empty def_star.

(* Two heaps agree if they are subset of a common heap *)

Definition agree h1 h2 := exists h1', exists h2', def (h1 :+ h1') /\ h1 :+ h1' = h2 :+ h2'. 

Lemma agree_refl h : def h -> agree h h.
Proof. by move=>h D; exists empty; exists empty; rewrite unh0. Qed.

Hint Resolve agree_refl.  

Lemma agreeC h1 h2 : agree h1 h2 <-> agree h2 h1.
Proof. by move=>h1 h2; split; move=>[h1'][h2'][D] E; exists h2'; exists h1'; rewrite -E. Qed.
      
Lemma agreehUnl t2 h t1 : agree h (t1 :+ t2) -> agree h t1. 
Proof. by move=>t2 h t1 [h1'][h2'][D] E; exists h1'; exists (t2 :+ h2'); rewrite unA. Qed.

Lemma agreehUnr t1 h t2 : agree h (t1 :+ t2) -> agree h t2.
Proof. by move=>t1 h t2; rewrite unC; move/agreehUnl. Qed.

Lemma agreeUnhl t2 h t1 : agree (t1 :+ t2) h -> agree t1 h.
Proof. by move=>t2 h t1; rewrite !(agreeC _ h); move/agreehUnl. Qed.

Lemma agreeUnhr t1 h t2 : agree (t1 :+ t2) h -> agree t2 h.
Proof. by move=>t1 h t2; rewrite !(agreeC _ h); move/agreehUnr. Qed.

Lemma agreeUnK t2 h2 t1 h1 : agree (t1 :+ t2) (h1 :+ h2) -> agree t1 h1.
Proof. by move=>t2 h2 t1 h1; move/agreeUnhl; move/agreehUnl. Qed.

Lemma agreeKUn t1 h1 t2 h2 : agree (t1 :+ t2) (h1 :+ h2) -> agree t2 h2.
Proof. by move=>t1 h1 t2 h2; move/agreeUnhr; move/agreehUnr. Qed.

(* Invertibility *)

Section Invertibility.
Variables (A : Type) (F : A -> Rel heap).
 
Definition invertible := 
  forall x1 x2 h1 h2, h1 \In F x1 -> h2 \In F x2 -> agree h1 h2 -> 
                        x1 = x2 /\ h1 = h2.

Definition invert := forall x1 x2 h, h \In F x1 -> h \In F x2 -> x1 = x2.

Definition proper (R : Rel heap) := forall h, h \In R -> def h.

Lemma iinv : invertible -> (forall x, proper (F x)) -> invert.
Proof.
move=>H P x1 x2 h H1 H2.
by case: (H _ _ _ _ H1 H2 (agree_refl (P x1 h H1))).
Qed.
End Invertibility.  

Lemma ptr_invert A p : invertible (fun x:A => p --> x).
Proof.
move=>A q x1 x2 h1 h2; do 2![case/swp=>-> _]; move=>[h1'][h2'][D].
by heap_cancel=>// _ _ ->.
Qed.

Lemma prod_invert A1 A2 (F1 : A1 -> Rel heap) (F2 : A2 -> Rel heap) : 
        invertible F1 -> invertible F2 -> 
          invertible (fun v => F1 v.1 # F2 v.2).
Proof.
move=>A1 A2 F1 F2 H1 H2 [v1 w1][v2 w2] h1 h2 /=.
move=>[i1][i2][->][G1][G2] D1 [r1][r2][->][W1][W2] D2 A.
move: (H1 _ _ _ _ G1 W1 (agreeUnK A)) (H2 _ _ _ _ G2 W2 (agreeKUn A)).
by move=>[->->][->->].
Qed.

(* SepIter over a finite set, and partitioning *)

(* Iteration over sequences *)
Module Iter.
Section General.
Variable A : Type.

Fixpoint iter (s : seq A) (f : A -> heap) {struct s} : heap :=
  if s is x::s' then f x :+ (iter s' f) else empty.

Fixpoint sepit (s : seq A) (f : A -> Rel heap) {struct s} : Rel heap :=
  if s is x::s' then f x # (sepit s' f) else emp.

Lemma it0 f : iter [::] f = empty.
Proof. by []. Qed.

Lemma it_cons x s f : iter (x::s) f = f x :+ iter s f.
Proof. by []. Qed.

Lemma it_cat s1 s2 f : iter (s1 ++ s2) f = iter s1 f :+ iter s2 f.
Proof.
elim=>[|x s1 IH] s2 f /=; first by rewrite un0h.
by rewrite IH unA.
Qed.

Lemma it_add_last s x f : iter (rcons s x) f = iter s f :+ f x.
Proof.
elim=>[|y s IH] x f /=; first by rewrite unh0 un0h.
by rewrite IH unA.
Qed.

Lemma sepit_def : forall s f, Def (sepit s f).
Proof. by case=>[|x sx] f //=. Qed.

Lemma sepit0 f : sepit [::] f = emp. Proof. by []. Qed.

Lemma sepit_cons x s f : sepit (x::s) f = f x # (sepit s f).
Proof. by []. Qed.

Lemma sepit_cat s1 s2 f : sepit (s1 ++ s2) f <=> sepit s1 f # sepit s2 f.
Proof.
elim=>[|x s1 IH] s2 f /=.
- by rewrite star0p.
by rewrite IH starA.
Qed. 

Lemma has_takedrop (p : pred A) (s : seq A) x :
        has p s ->
          s = take (find p s) s ++ [::nth x s (find p s)] ++ drop (find p s).+1 s.
Proof.
move=>p s x H.
set i := find p s.
have L1 : i < size s by rewrite /i -has_find.
rewrite -{1}(cat_take_drop i s).
by rewrite (drop_nth x).
Qed.
End General.

Section EqualityTp.

Variable (A : eqType).

Lemma mem_split (x : A) (s : seq A) :
        x \in s -> exists s1, exists s2, s = s1 ++ [:: x] ++ s2.
Proof.
move=>x s H.
have L : has (pred1 x) s.
- by case: hasP=>//; elim; exists x=>/=.
move: (has_takedrop x L).
set i := find (pred1 x) s.
rewrite (_ : nth x s i = x); last first.
- suff : pred1 x (nth x s i) by move/eqP.
  by apply: nth_find.
eauto.
Qed.

Lemma mem_uniq_spli (x : A) (s : seq A) :
       x \in s -> uniq s ->
         exists s1, exists s2, s = s1 ++ [:: x] ++ s2 /\
           uniq (s1 ++ s2) /\ x \notin (s1 ++ s2).
Proof.
move=>x s H1 H2; move: (mem_split H1)=>[s1 [s2 H3]].
exists s1; exists s2; move: H2.
rewrite H3 uniq_catCA cons_uniq.
by move/andP=>[].
Qed.

Lemma it_perm_eq (s1 s2 : seq A) f : perm_eq s1 s2 -> iter s1 f = iter s2 f.
Proof.
elim=>[|x s1 IH] s2 f /=.
- by move/perm_eq_size; move/eqP; rewrite eq_sym; move/eqP; move/size0nil=>->.
move=>H.
have L1: has (pred1 x) s2.
- rewrite has_count; case: perm_eqP H=>//; move/(_ (pred1 x))=><-.
  by rewrite /= eq_refl.
have L2: x \in s2.
- by case: hasP L1=>//= [[y]] H1; move/eqP=><-.
move: (mem_split L2)=>[t1][t2] E.
rewrite E in H *.
have L3: perm_eq (x::s1) ([:: x] ++ t1 ++ t2). 
- apply: (perm_eq_trans H).
  by rewrite perm_catCA.
rewrite /= perm_cons in L3.
by move: (IH (t1 ++ t2) f L3)=>->; rewrite !it_cat /=  unh0 unCA.
Qed.

Lemma filter_predU p (s : seq A) : 
        s =i filter p s ++ filter (predC p) s.
Proof.
move=>p s x.
rewrite -{1}(filter_predT s).
rewrite mem_cat !mem_filter /=.
case E: (p x)=>//=.
by rewrite orbF.
Qed.

Lemma sepit_inv (s : seq A) f1 f2 h1 h2 : 
        h1 \In sepit s f1 -> h2 \In sepit s f2 -> agree h1 h2 ->
        (forall k h1 h2, k \in s -> h1 \In f1 k -> h2 \In f2 k -> agree h1 h2 -> h1 = h2) ->
        h1 = h2.
Proof.
elim=>[|x s IH] f1 f2 h1 h2 /=; first by case=>->->.
move=>[r1][t1][->][H1][S1] D1[r2][t2][->][H2][S2] D2 L H.
have S : x \in x :: s. by rewrite inE eq_refl.
rewrite (H x r1 r2 S H1 H2 (agreeUnK L)); heap_congr.
apply: (IH _ _ _ _ S1 S2 (agreeKUn L))=>k h3 h4 E.
by apply: H; rewrite inE E orbT.
Qed.

Lemma sepitP (x : A) (s : seq A) f : uniq s ->
       sepit s f =r if x \in s then f x # sepit (filter (predC1 x) s) f
                    else sepit s f.
Proof.
move=>x s; elim: s x=>[|y s IH] x f //=.
rewrite inE; case: eqP=>/==>[->|H1].
- rewrite eq_refl /=; case/andP=>H1 H2.
  rewrite (IH y f H2) (negbTE H1). 
  have filter_predC1 : forall (x:A) s,
    x \notin s -> filter (predC1 x) s = s.
  - move=>x'; elim=>[|y' s' IH'] //.
    rewrite inE; case: eqP=>//=; case: eqP=>[->|] //= _ _ H.
    by rewrite IH'.
  by rewrite filter_predC1.
case/andP=>H2 H3; case: eqP H1=>[->| _ _] //.
case: ifP=>H4 //=.
by rewrite starCA (IH x f H3) H4.
Qed.

Lemma sepit_emp (s : seq A) f : (forall x, x \in s -> f x =r emp) -> sepit s f =r emp.
Proof.
elim=>[|x s IH] f //= H; rewrite (H x) ?inE ?eq_refl //.
apply: defrel=>//; rewrite star0p IH // => y S.
by apply: H; rewrite inE S orbT.
Qed.

End EqualityTp.

(* Special case when I is a finite type *)

Lemma perm_eq_sepit (I : finType) (s1 s2 : seq I) f : 
        perm_eq s1 s2 -> Iter.sepit s1 f =r Iter.sepit s2 f.
Proof.
move=>I.
elim=>[|x s1 IH] s2 f /=.
- by move/perm_eq_size; move/eqP; rewrite eq_sym; move/eqP; move/size0nil=>->.
move=>H.
have L1: has (pred1 x) s2.
- rewrite has_count; case: perm_eqP H=>//; move/(_ (pred1 x))=><-.
  by rewrite /= eq_refl.
have L2: x \in s2.
- by case: hasP L1=>//= [[y]] H1; move/eqP=><-.
move: (mem_split L2)=>[t1][t2] E.
rewrite E in H *.
have L3: perm_eq (x::s1) ([:: x] ++ t1 ++ t2). 
- apply: (perm_eq_trans H).
  by rewrite perm_catCA.
rewrite /= perm_cons in L3.
move: (IH (t1 ++ t2) f L3)=>->.
apply: defrel=>//; last by apply: sepit_def.
by rewrite !sepit_cat /= starp0 starCA.
Qed.

Lemma eq_sepitF (I : finType) (s : seq I) f1 f2 : 
        (forall x, x \in s -> f1 x =r f2 x) -> sepit s f1 =r sepit s f2.
Proof.
move=>I; elim=>[|x s IH] f1 f2 //= H.
rewrite (IH f1 f2) //; first by rewrite (H x) // inE eq_refl.
by move=>y H1; apply: H; rewrite inE H1 orbT.
Qed.

End Iter.

(* If I is a finite type, we can work with subsets, instead of sequences *)
(* to mod out by ordering and repetition *)

Section FinIter.
Variable I : finType.

Definition seq_of (s : {set I}) := filter (pred_of_set s) (enum I).

Definition sepit (s : {set I}) (Ps : I -> Rel heap) := 
  Iter.sepit (seq_of s) Ps.

Lemma sepit_def : forall s f, Def (sepit s f).
Proof. by move=>s f; apply: Iter.sepit_def. Qed.

Lemma setE (p : pred I) : pred_of_set (finset p) =1 p.
Proof.
move=>p x.
rewrite SetDef.pred_of_setE unlock /fgraph SetDef.finsetE FunFinfun.finfunE. 
rewrite (tuple.tnth_nth false) (nth_map x).
- by rewrite nth_index // mem_enum.
by rewrite index_mem mem_enum.
Qed.

Lemma setq (s1 s2 : {set I}) : s1 = s2 <-> seq_of s1 =i seq_of s2.
Proof.
move=>s1 s2; split=>[->|H] //; rewrite -setP=>x; move: (H x). 
by rewrite /seq_of !mem_filter -enumT mem_enum !andbT /in_mem.
Qed.

Lemma seqt (s1 s2 : seq I) : s1 =i s2 <-> 
                             finset (topred s1) =i finset (topred s2).
Proof.
move=>s1 s2; split=>H x; first by rewrite !in_set !topredE H.
by move: (H x); rewrite !in_set /in_mem.
Qed.

Lemma mem_seq_of (s : {set I}) x : (x \in seq_of s) = (x \in s).
Proof. by move=>s x; rewrite /seq_of mem_filter mem_enum andbT. Qed.

Lemma uniq_seq_of (s : {set I}) : uniq (seq_of s).
Proof. by move=>s; rewrite /seq_of filter_uniq // enum_uniq. Qed. 

Lemma perm_eq_seqP (s1 s2 : {set I}) : 
       reflect (seq_of s1 =i seq_of s2) 
               (perm_eq (seq_of s1) (seq_of s2)).
Proof.
move=>s1 s2; case E: (perm_eq _ _); constructor.
- by apply: perm_eq_mem.
by move=>H; elim: (elimF idP E); apply: uniq_perm_eq H; rewrite uniq_seq_of.
Qed.

Lemma eq_sepit (s1 s2 : {set I}) f : s1 =i s2 -> sepit s1 f =r sepit s2 f.
Proof.
move=>s1 s2 f; rewrite setP setq. 
suff: perm_eq (seq_of s1) (seq_of s2) -> sepit s1 f =r sepit s2 f.
- by move=>H H2; apply: H; case: perm_eq_seqP.
by apply: Iter.perm_eq_sepit.
Qed.

Lemma sepit0 f : sepit set0 f =r emp.
Proof. 
move=>f; rewrite /sepit /seq_of.
rewrite (Iter.perm_eq_sepit (s2 := filter pred0 (enum I))); first by rewrite filter_pred0.
apply: uniq_perm_eq; try by rewrite filter_uniq // enum_uniq.
by move=>x; rewrite !mem_filter setE.
Qed.

Lemma sepit_emp (s : {set I}) f : (forall x, x \in s -> f x =r emp) -> sepit s f =r emp.
Proof.
move=>s f H; rewrite /sepit; apply: Iter.sepit_emp=>x.
by rewrite mem_seq_of; apply: H.
Qed.

Lemma sepitS x (s : {set I}) f : 
        sepit s f =r if x \in s then f x # sepit (s :\ x) f 
                     else sepit s f.
Proof.
move=>x s f.
apply: (releq_trans (Iter.sepitP x (s:=seq_of s) f (uniq_seq_of s))).
rewrite mem_seq_of.
case E: (x \in s) =>//.
rewrite (@Iter.perm_eq_sepit _ (filter (predC1 x) (seq_of s)) (seq_of (s :\ x))) //.
rewrite /seq_of -filter_predI.
apply: uniq_perm_eq=>[||y]; try by rewrite filter_uniq // enum_uniq.
by rewrite !mem_filter setE in_set.
Qed.

(* Some frequently use special instances if eq_sepitS *)

Lemma sepitU1 x (s : {set I}) f : sepit (x |: s) f =r f x # sepit (s :\ x) f.
Proof.
by move=>x s f; rewrite (sepitS x) setU11 setDUl setDv set0U.
Qed.

Lemma sepitT1 x f : sepit setT f =r f x # sepit (setT :\ x) f.
Proof. by move=>x f; rewrite (sepitS x) in_setT. Qed.

Lemma sepitF (s : {set I}) f1 f2 : 
        (forall x, x \in s -> f1 x =r f2 x) -> sepit s f1 =r sepit s f2.
Proof.
move=>s f1 f2 H; apply: Iter.eq_sepitF=>x H1; apply: H.
by rewrite -mem_seq_of.
Qed.

Lemma sepit_invert (s : {set I}) f1 f2 h1 h2 : 
        h1 \In sepit s f1 -> h2 \In sepit s f2 -> agree h1 h2 ->
        (forall k h1 h2, k \in s -> h1 \In f1 k -> h2 \In f2 k -> agree h1 h2 -> h1 = h2) ->
        h1 = h2.
Proof.
move=>s f1 f2 h1 h2 H1 H2 H3 H4; apply: (Iter.sepit_inv H1 H2 H3).
move=>k i1 i2 H5; apply: H4.
by rewrite mem_seq_of in H5.
Qed.

End FinIter.

Implicit Arguments setq [I s1].
Implicit Arguments seqt [I s1].
Implicit Arguments eq_sepit [I s1 f].
Implicit Arguments sepitF [I s f1].

(*********************************************************************)
(* Now a section with useful additional lemmas over finite types and *)
(* ordinals.                                                         *)
(*********************************************************************)

Section Ordinals.

Lemma paths_take (s : seq nat) i x : path leq x s -> path leq x (take i s).
Proof. by elim=>[|z s IH] //=; case=>[|i] x //=; case/andP=>-> H; apply: IH. Qed.

Lemma paths_drop (s : seq nat) i x : path leq x s -> path leq x (drop i s).
Proof.
elim=>[|z s IH] //; case=>[|i] //= x; case/andP=>H1 H2; apply: IH.
case: s H2=>[|s xs] //=; case/andP=>H2 ->; rewrite andbT.
by apply: leq_trans H2.
Qed.

Lemma sorted_take s i : sorted leq s -> sorted leq (take i s).
Proof. by elim=>[|z s IH] //; case=>[|i] //; apply: paths_take. Qed.

Lemma sorted_drop s i : sorted leq s -> sorted leq (drop i s).
Proof. by elim=>[|z s IH] //; case=>[|i] //= H; apply: IH; apply: path_sorted H. Qed.

Section Uniq.
Variable (A : eqType).

Lemma uniq_take (s : seq A) i : uniq s -> uniq (take i s).
Proof. 
elim=>[|x s IH] //; case=>[|i] //=; case/andP=>H1 H2; rewrite (IH _ H2) andbT.
by apply: contra H1; apply: mem_take.
Qed.

Lemma uniq_drop (s : seq A) i : uniq s -> uniq (drop i s).
Proof. by elim=>[|x s IH] //; case=>[|i] //=; case/andP=>H1 H2; apply: IH. Qed.
End Uniq.

(* starting from m, iota generates i+k successive numbers *)
Lemma take_iota_index i k m : take i (iota m (i + k)) = iota m i.
Proof. by elim=>[|i IH] k m; rewrite ?take0 // addSn /= IH. Qed.

Lemma drop_iota_index i k m : drop i (iota m (i + k)) = iota (m + i) k.
Proof.
elim=>[|i IH] k m; first by rewrite drop0 addn0 add0n.
by rewrite addSn /= IH addSnnS. 
Qed.

Lemma take_iota0 i n : take i (iota 0 n) = if i <= n then iota 0 i else iota 0 n.
Proof.
move=>i n; case: ifP=>H; first by rewrite -(subnKC H) take_iota_index.
rewrite take_oversize // size_iota.
by rewrite leq_eqVlt; case: ltnP; rewrite ?orbT ?H.
Qed.

Lemma drop_iota0 i n : drop i (iota 0 n) = if i <= n then iota i (n - i) else [::].
Proof.
move=>i n; case: ltnP=>H; last by rewrite -(subnKC H) drop_iota_index add0n subnKC.
by rewrite drop_oversize // size_iota leq_eqVlt H orbT.
Qed.

End Ordinals. 

Definition indx := fun (I : finType) (x : I) => index x (enum I).
Implicit Arguments indx [I].
Prenex Implicits indx.

Section Enums.
Variable (I : finType).

Lemma size0F : forall A, 0 < size (Nil A) -> False.
Proof. by []. Qed.

Lemma card0P : 0 < #|I| -> size (enum I) > 0.
Proof. by rewrite cardE. Qed. 

Lemma cardiP i : i < #|I| -> 0 < #|I|. 
Proof. by move=>i H; apply: leq_trans H. Qed.

Definition dflt (pf : 0 < #|I|) : I :=
  match (enum I) as l return 0 < size l -> I with
    x :: _ => fun pf => x
  | [::] => fun pf => False_rect I (size0F pf)
  end (card0P pf).

Definition ith i (pf : i < #|I|) : I := nth (dflt (cardiP pf)) (enum I) i.

Lemma indx_ith i (pf : i < #|I|) : indx (ith pf) = i.
Proof.
by move=>i pf; rewrite /ith /= /indx index_uniq // -?cardE // enum_uniq.
Qed.

Lemma ith_indx s (pf : indx s < #|I|) : ith pf = s.
Proof. by move=>s /= pf; rewrite /ith /= nth_index // mem_enum. Qed.   

Lemma indx_inj : injective (@indx I). 
Proof. 
rewrite /index /indx /= =>x1 x2.
have [H1 H2] : x1 \in (enum I) /\ x2 \in (enum I) by rewrite !mem_enum.
elim: (enum I) H1 H2 =>[|y s IH]  //=; rewrite !inE.
case: ifP=>H1.
- rewrite (eqP H1) eq_refl /= => _.
  case: ifP=>H2; by rewrite ?(eqP H2) // eq_sym H2.
rewrite eq_sym H1 /= => T1.
by case: ifP=>H2 //; rewrite eq_sym H2 /= => T2; case; apply: IH.
Qed.

Lemma indx_injE s i (pf : i < #|I|) : (s == ith pf) = (indx s == i).
Proof.
move=>s i pf; case: eqP=>[->|]; first by rewrite indx_ith eq_refl.
by case: eqP=>[H|] //; move: H pf=><- pf; rewrite ith_indx.
Qed.

Lemma map_indx : map indx (enum I) = iota 0 #|I|.
Proof.
rewrite cardE /indx /index; move: (enum_uniq I).
elim: (enum I)=>[|x xs IH] //=; case/andP=>H1 H2; rewrite eq_refl.
rewrite (@eq_in_map _ _ _ (fun x0 => (index x0 xs).+1)); last first.
- by move=>z R /=; case: ifP H1=>R1 //; rewrite (eqP R1) R.
by congr (0 :: _); rewrite (_ : 1 = 1 + 0) // iota_addl -(IH H2) -map_comp.
Qed.
 
Lemma take_enum x i : x \in take i (enum I) = (indx x < i).
Proof.
move=>x i; pose f x := indx x.
rewrite -(mem_map indx_inj) map_take map_indx take_iota0.
case: (leqP i #|I|)=>H; rewrite mem_iota /=; first by rewrite add0n.
rewrite add0n cardE index_mem mem_enum inE /=; apply: sym_eq.
apply: (@ltn_trans #|I|) H.
by rewrite cardE index_mem mem_enum.
Qed.

Lemma drop_enum x i : x \in drop i (enum I) = (i <= indx x).
Proof.
move=>x i; pose f x := index x (enum I).
rewrite -(mem_map indx_inj) map_drop map_indx drop_iota0.
case: ifP=>H; first by rewrite mem_iota subnKC // cardE index_mem mem_enum andbT.
apply: sym_eq; apply: negbTE; move/negbT: H; apply: contra=>H.
by apply: (leq_trans H); rewrite cardE index_size.
Qed.

Lemma take_enum_filter k : 
        filter (preim indx [pred x | x < k]) (enum I) = take k (enum I).
Proof.
move=>k; apply: (inj_map indx_inj).
rewrite map_take map_indx -filter_map map_indx.
apply: (eq_sorted leq_trans anti_leq).
- by apply: (sorted_filter leq_trans); apply: sorted_iota.
- by apply: sorted_take; apply: sorted_iota.
apply: uniq_perm_eq; rewrite ?filter_uniq // ?uniq_take // ?iota_uniq //.
move=>x; rewrite mem_filter take_iota0 cardE.
case: (leqP k (size _))=>H1; rewrite !mem_iota !add0n /=.
- by case: (ltnP x k)=>H2 //=; apply: leq_trans H1.
case: (ltnP x (size _))=>H2 //=; last by rewrite andbF.
by rewrite andbT; apply: ltn_trans H1.
Qed.

Lemma drop_enum_filter k : 
       filter (preim indx [pred x | x >= k]) (enum I) = drop k (enum I).
Proof.
move=>k; apply: (inj_map indx_inj).
rewrite map_drop map_indx -filter_map map_indx.
apply: (eq_sorted leq_trans anti_leq).
- by apply: (sorted_filter leq_trans); apply: sorted_iota.
- by apply: sorted_drop; apply: sorted_iota.
apply: uniq_perm_eq; rewrite ?filter_uniq ?uniq_drop // ?iota_uniq //.
move=>x; rewrite mem_filter drop_iota0 cardE.
case: (leqP k (size _))=>H1; rewrite !mem_iota add0n /=. 
- by rewrite subnKC.
case: ltngtP=>//; rewrite ?andbF ?andbT //.
move=>H2; apply: negbTE; rewrite leqNgt negbK.
by apply: ltn_trans H1.
Qed.

End Enums.

Section EnumOrds.
Variable n : nat.

Lemma take_enum_ord (x : 'I_n) (i : nat) :
       (x \in take i (enum 'I_n)) = (eqtype.val x < i).
Proof. by move=>x i; rewrite take_enum /= /indx index_enum_ord. Qed.

Lemma drop_enum_ord (x : 'I_n) (i : nat) :
           (x \in drop i (enum 'I_n)) = (i <= eqtype.val x).
Proof. by move=>x i; rewrite drop_enum /= /indx index_enum_ord. Qed.

Lemma take_size_ord : take n (enum 'I_n) = (enum 'I_n).
Proof. by rewrite take_oversize // size_enum_ord. Qed.

Lemma drop_size_ord m : n <= m -> drop m (enum 'I_n) = [::].
Proof. by move=>m H; rewrite drop_oversize // size_enum_ord. Qed.

End EnumOrds.










