Require Import ssreflect ssrbool ssrnat eqtype seq ssrfun.
Require Import rels heaps stmodel stsep hprop vcgen.
Set Implicit Arguments. 
Unset Strict Implicit. 
Import Prenex Implicits.

(* Linked lists, storing a value and next pointer in consecutive locations *)

Definition llist (T : Type) := loc.
 
Section LList.
Variable T : Type.
Notation llist := (llist T).

Fixpoint lseg (p q : loc) (xs : seq T) {struct xs} := 
  if xs is x::xt then 
      [Rel h | exists r, exists h', [/\ def (p :-> x :+ (p .+ 1 :-> r :+ h')), 
                 p :-> x :+ (p .+ 1 :-> r :+ h') = h & h' \In lseg r q xt]]
  else [Rel h | p = q /\ h = empty].

Lemma lseg_add_last1 xs x p r h : 
        h \In lseg p r (rcons xs x) ->
        exists q, exists h', [/\ def (h' :+ (q :-> x :+ q .+ 1 :-> r)), 
          h' :+ (q :-> x :+ q .+ 1 :-> r) = h & h' \In lseg p q xs].
Proof.
elim=>[x p r h [q][h'][D] <- [<- ->]|y xs IH x p r h [t][i][D] <- H]/=.
- by hauto D; rewrite unh0 un0h. 
case/IH: H D=>q [h'][_][<-] H D.
exists q; exists (p :-> y :+ (p .+ 1 :-> t :+ h')); hauto D.
by rewrite !unA.
Qed.

Lemma lseg_add_last xs x p r h : 
        (exists q, exists h', 
          [/\ def (h' :+ (q :-> x :+ q .+ 1 :-> r)), 
              h' :+ (q :-> x :+ q .+ 1 :-> r) = h & h' \In lseg p q xs])-> 
          h \In lseg p r (rcons xs x).
Proof.
elim=>[|x xs IH] y p r h.
- by move=>[q][h'][D][<-][->->]; hauto D; rewrite unh0 un0h.
move=>[q][h'][D][<-] E; move: E D=>/= [z][h''][D2] /= [<-] H1 D.
by rewrite -!unA; hauto D; apply: IH; hauto D.
Qed. 

Lemma lseg_def p q xs : proper (lseg p q xs).
Proof. 
by move=>p q xs; case: xs p q=>[p q i [_ ->]|x xs p q i [next][h][D][<-]]. 
Qed.

Lemma lseg_null xs q h : h \In lseg null q xs -> [/\ q = null, xs = [::] & h = empty].
Proof. 
by case=>[|x xs] q h; [case=>->-> | case=>next[h'][]; rewrite defPUn eq_refl].
Qed.

Lemma lseg_neq xs p q h : p != q -> h \In lseg p q xs -> 
        exists x, exists r, exists h', 
          [/\ xs = x :: (behead xs), def (p :-> x :+ (p .+ 1 :-> r :+ h')), 
              p :-> x :+ (p .+ 1 :-> r :+ h') = h & h' \In lseg r q (behead xs)].
Proof.
case=>[|x xt] p q h /= H []; first by move=>E; rewrite E eq_refl in H.
by move=>y [h'][D]<- H1; hauto D.
Qed.

Lemma lseg_empty xs p q : empty \In lseg p q xs -> p = q /\ xs = [::].
Proof.
case=>[|x xs] p q /=; first by case.
by case=>r [h]; case=>_; case/unhh0E; move/emphT; rewrite emp_pt.
Qed.

Lemma lseg_case xs p q h : h \In lseg p q xs -> 
        [/\ p = q, xs = [::] & h = empty] \/
        exists x, exists r, exists h',
        [/\ xs = x :: behead xs, h = p :-> x :+ (p .+ 1 :-> r :+ h') & 
            h' \In lseg r q (behead xs)].
Proof.
case=>[|x xs] p q h /=; first by case=>->->; left.
by case=>r[h']; case=> D <- => H; right; hauto D.
Qed.

Lemma lseg_inv : forall p q, invert (lseg p q).
Proof.
move=>p q xs1; elim: xs1 p q=>[|x1 xs1 IH] p q xs2 h /=.
- by case=>->->; case/lseg_empty.
move=>[q1][r1][D1][<-] H1; elim: xs2=>/= [[_]|x2 xs2 Ih2 [x][h'][D]].
- by case/unhh0E; move/emphT; rewrite emp_pt.
by heap_cancel=>// _ ->->->; move/(IH _ _ _ _ H1)=>->.
Qed.

(* Special case when p = null *)
Definition lseq p := lseg p null.

Lemma lseq_def p xs : proper (lseq p xs). 
Proof. by move=>p; apply: lseg_def. Qed.

Lemma lseq_null xs h : h \In lseq null xs -> xs = [::] /\ h = empty.
Proof. by move=>xs h; case/lseg_null=>_ ->. Qed.

Lemma lseq_pos xs p h : p != null -> h \In lseq p xs -> 
        exists x, exists r, exists h', 
          [/\ xs = x :: (behead xs), def (p :-> x :+ (p .+ 1 :-> r :+ h')), 
              p :-> x :+ (p .+ 1 :-> r :+ h') = h & h' \In lseq r (behead xs)].
Proof. by move=>xs p h; apply: lseg_neq. Qed.

Lemma lseq_invert : forall p, invertible (lseq p).
Proof.
rewrite /lseq=>p xs1; elim: xs1 p=>[|x1 xs1 IH] p xs2 h1 h2 /=.
- by case=>->->; case/lseg_null=>_ ->->.
move=>[q1][i1][D] <- H1; case/lseg_case=>[[->]->->|[p2][r2][i2][->] -> H2].
- by case=>?[?][]; rewrite -unA defNf. 
rewrite !unA=>R; move: (agreeUnK R) (agreeKUn R) H1 H2=>{R} [s1][s2][D2].
by heap_cancel=>// _ _ <-<- R H1 H2; case: (IH _ _ _ _ H1 H2 R)=>->->.
Qed.

Lemma lseq_inv p : invert (lseq p).
Proof. by move=>p; apply: iinv; [apply: lseq_invert | apply: lseq_def]. Qed.

Program 
Definition insert p x : STsep loc (fun i => exists xs, i \In lseq p xs, 
                                   fun y i m => forall xs, i \In lseq p xs -> 
                                     exists q, m \In lseq q (x::xs) /\ y = Val q) :=
  Do (q <-- Allocb p 2; 
      q ::= x;;
      ret q). 
Next Obligation.
apply: {H} (ghost H) H0=>xs H; move: (lseq_def H); hstep=>q; heval; rewrite un0h; hhauto.
Qed.

Program 
Definition remove p : STsep loc (fun i => exists xs, i \In lseq p xs,
                                 fun y i m => forall xs, i \In lseq p xs -> 
                                   exists q, m \In lseq q (behead xs) /\ y = Val q) :=
  Do (If p == null then ret p 
      else pnext <-- !p .+ 1;
           dealloc p;; dealloc p .+ 1;;
           ret pnext). 
Next Obligation.
apply: (ghost H) H0=>xs H0; move: H0 (lseq_def H0). 
case: ifP=>H1; first by rewrite (eqP H1); case/lseq_null=>->->; heval.
by case/(lseq_pos (negbT H1))=>x [q][h][->] D <- /= H2; heval.
Qed.

(* let's try the full explicit proofs, without use of heval *)

Program 
Definition insert' p x : STsep loc (fun i => exists xs, i \In lseq p xs, 
                                   fun y i m => forall xs, i \In lseq p xs -> 
                                     exists q, m \In lseq q (x::xs) /\ y = Val q) :=
  Do (q <-- Allocb p 2; 
      q ::= x;;
      ret q). 
Next Obligation.
apply: (ghost H) H0=>xs H1. 
apply: bnd_allocb (lseq_def H1)=>//= q; rewrite -unA unh0.
by apply: bnd_write=>D; apply: val_ret=>//; do !econstructor.
Qed.

Program 
Definition remove' p : STsep loc (fun i => exists xs, i \In lseq p xs,
                                 fun y i m => forall xs, i \In lseq p xs -> 
                                   exists q, m \In lseq q (behead xs) /\ y = Val q) :=
  Do (If p == null then ret p 
      else pnext <-- !p .+ 1;
           dealloc p;; 
           dealloc p .+ 1;;
           ret pnext). 
Next Obligation.
apply: (ghost H) H0=>xs {H} H; case: ifP H (lseq_def H)=>H1.
- by rewrite (eqP H1); case/lseq_null=>->->; apply: val_ret; exists null.
case/(lseq_pos (negbT H1))=>x [q][h][->] _ <- /= H.
rewrite unCA=>D; apply: bnd_read=>//. 
rewrite unCA in D *; apply: bnd_dealloc=>{D} //.
by apply: bnd_dealloc; apply: val_ret; exists q.
Qed.

Definition shape_rev p s := lseq p.1 s.1 # lseq p.2 s.2.

Lemma shape_revD p : forall s, proper (shape_rev p s).
Proof. by move=>[p1 p2] [s1 s2] h [h1][h2][->] [] _ []. Qed.  

(*
Lemma rev_invert p : invertible (shape_rev p).
Proof. by move=>p; apply: prod_invert; apply: lseq_invert. Qed.

Lemma rev_inv p : invert (shape_rev p).
Proof. by move=>p; apply: iinv=>[|[??]?[?][?][_][_][_][]] //; exact: rev_invert. Qed.
*)

Definition revT : Type := 
  forall p, STsep loc (fun i => exists ps, i \In shape_rev p ps,
                       fun y i m => forall ps, i \In shape_rev p ps -> 
                         exists r, m \In lseq r (rev ps.1 ++ ps.2) /\ y = Val r). 

Program 
Definition reverse p : STsep loc (fun i => exists xs, i \In lseq p xs, 
                                  fun y i m => forall xs, 
                                    i \In lseq p xs -> 
                                    exists q, m \In lseq q (rev xs) /\ y = Val q) :=        
  Do (Fix (fun (reverse : revT) p => 
            (Do (If p.1 == null then ret p.2 
                 else xnext <-- !p.1 .+ 1;
                      p.1 .+ 1 ::= p.2;;
                      reverse (xnext, p.1)))) (p, null)).
Next Obligation. 
apply: (ghost (s, s0)) H0=>[[x1 x2]][i1][i2][->] /=.
move: l l0 {s s0}=>p1 p2 [H1] [H2].
case: eqP H1=>[->|H]; first by case/lseq_null=>->->; rewrite un0h; heval.
case/lseq_pos=>[|xd [xn][h'][->] D <- /= H1]; first by case: eqP. 
do 2!hstep; rewrite -!(unCA h') -(pull p1)=>D1. 
apply: (cons_gh (behead x1, xd::x2))=>[??[?][?][->]|??[?][?]||] //; hauto D1.
by rewrite rev_cons cat_rcons.
Qed.
Next Obligation.
apply: (ghost H) H0=>{H} xs H; move: (lseq_def H)=>H0. 
apply: (cons_gh (xs, Nil T))=>[x h [r] /= [H1][->]|??[?][]||] //. 
- by rewrite cats0 in H1; eauto. 
by exists i; hhauto.
Qed.

End LList. 

Prenex Implicits lseg lseq.

