Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq.
Require Import rels heaps hprop stmodel stsep vcgen llist.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Record queue (T : Type) : Type := Queue {front: loc; back: loc}. 
Definition EmptyQueue : exn := exn_from_nat 100.

Module Queue. Section Queue.
Variable T : Type.
Notation queue := (queue T).

Definition shape' (fr bq : loc) (xs : seq T) :=
  if fr == null then [Rel h | [/\ bq = null, xs = [::] & h = empty]]
  else [Rel h | exists xt, exists x, exists h',  
        [/\ xs = rcons xt x, def (bq :-> x :+ (bq .+ 1 :-> null :+ h')), 
            (bq :-> x :+ (bq .+ 1 :-> null :+ h')) = h & h' \In lseg fr bq xt]].

Definition shape (q : queue) (xs : seq T) := 
  [Rel h | exists fr, exists bq, exists h',
    [/\ def (front q :-> fr :+ (back q :-> bq :+ h')), 
            (front q :-> fr :+ (back q :-> bq :+ h')) = h &
            h' \In shape' fr bq xs]].

Lemma shape'_nil fr bq h : 
        h \In shape' fr bq [::] -> [/\ fr = null, bq = null & h = empty].
Proof.
by move=>fr bq h; rewrite /shape'; case: eqP=>[->[->]_->| N [[|y xt]][x][h'][]].
Qed.

Lemma shape'_add_last fr bq xt x h : 
         (exists h', [/\ def (bq :-> x :+ (bq .+ 1 :-> null :+ h')), 
                         bq :-> x :+ (bq .+ 1 :-> null :+ h') = h &
                         h' \In lseg fr bq xt]) -> 
         h \In shape' fr bq (rcons xt x).
Proof.
rewrite /shape'=>fr bq xt x h [h'][D] <- H.
by case: eqP H D=>[->|_ H D]; [case/lseg_null=>->->->; rewrite defNf | hauto D].
Qed.

Lemma backfront fr bq xs h : 
        h \In shape' fr bq xs -> (fr == null) = (bq == null).
Proof.
by rewrite /shape'=>fr bq xs h; case: ifP=>[E [->]_ _| E [xt][x][h'][_] D];
   rewrite E ?eq_refl //; case: eqP D=>// ->; rewrite defNf.
Qed.

Lemma shapeD q xs : proper (shape q xs).
Proof. move=>q xs h [h1] [bq][h'] [] D <-. by []. Qed.    

Program Definition new : 
          STsep queue (emp, fun y i h => exists v, y = Val v /\ h \In @shape v [::]) :=
  Do (x <-- alloc null; 
      y <-- alloc null;
      ret (Queue T x y)).
Next Obligation.
by case: H emptyD=><-; hstep=>x; hstep=>y; do !hstep; rewrite unCA; hhauto.
Qed. 

Program Definition free (q : queue) : 
          STsep unit (shape q [::], fun y i h => y = Val tt /\ h = empty) :=
  Do (dealloc (front q);;
      dealloc (back q)).
Next Obligation. 
by case: H2=>[D]<- H3; move: D; do !hstep; case/shape'_nil: H3=>_ _ ->. 
Qed. 

Program Definition enq (q : queue) (x : T) : 
  STsep unit (fun i => exists xs, i \In shape q xs, 
              fun y i h => forall xs, i \In shape q xs -> y = Val tt /\
                                      h \In shape q (rcons xs x)) :=
  Do (next <-- Allocb null 2;
      next ::= x;;
      ba <-- !back q; 
      back q ::= next;;
      If ba == null then front q ::= next
      else ba .+ 1 ::= next).
Next Obligation.
apply: (ghost H) H0=>{H} xs [fr][bq][h']{H0}[D][<-] H; move: D.
hstep=>next; do !hstep; rewrite -(backfront H) un0h. 
case: ifP H=>Ef; rewrite /shape' ?Ef.
- by case=>_->->; do !hstep=>D; hauto D; apply: shape'_add_last; hauto D.
case=>[s2][x2][i2][->] D <- H2; do !hstep. 
rewrite -!(push (back q)) -!(push (front q)) 4!(push (bq .+ 1))=>D1.
hauto D1; apply: shape'_add_last; hauto D1; apply: lseg_add_last.
by rewrite -!(unCA i2); hauto D1.
Qed.  

Program Definition deq (q : queue) : 
  STsep T (fun i => exists xs, shape q xs i, 
           fun y i h => forall xs, shape q xs i -> shape q (behead xs) h /\
             if y is Val v then xs = v :: behead xs  
             else y = Exn EmptyQueue /\ xs = [::]) :=
  Do (fr <-- !front q;
      If fr == null then Throw _ EmptyQueue
      else 
        x <-- !fr;
        next <-- !fr .+ 1;
        front q ::= next;;
        dealloc fr;;
        dealloc fr .+ 1;;
        If next == null then back q ::= null;;
          ret x
        else ret x).
Next Obligation.
apply: (ghost H); last by eauto.
move=>{H H0 H1 H2 H3 H4}=>xs [fr][bq][h][D][<-] H; move: D; do !hstep.
case: ifP H=>Ef; rewrite /shape' Ef.
- by case=>->->->; heval=>D; hauto D; rewrite Ef; hauto D.
case=>[[|y xt]][x][h'][->] D <-. 
- by case=>->->; do !hstep; rewrite eq_refl; do !hstep; rewrite unCA=>D1; hauto D1.
case=>next [h2] [D1] <- H1; do !hstep; case: ifP H1 D1 D=>En; rewrite ?(eqP En). 
- by case/lseg_null=>->->-> _; rewrite defNf.
by move=>H1 D1 D; heval=>D2; hauto D2; rewrite En /=; hauto D2.
Qed.

End Queue. End Queue.